from liblaf import grapes

with grapes.optional_imports(extra="duration"):
    import about_time


def human_count(count: int, unit: str = "", prec: int | None = None) -> str:
    # TODO: remove dependency on `about-time`
    hc = about_time.HumanCount(count, unit)
    return hc.as_human(prec)



import logging
from collections.abc import Sequence

import loguru

from . import init_icecream, init_loguru, init_rich


def init_logging(
    level: int | str = logging.NOTSET,
    *,
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
    traceback_show_locals: bool = True,
) -> None:
    init_rich(show_locals=traceback_show_locals)
    init_loguru(level=level, handlers=handlers, levels=levels)
    init_icecream()



from collections.abc import Generator, Iterable
from typing import Any, TypeVar

from rich.progress import Progress

from liblaf import grapes

from . import progress

_T = TypeVar("_T")


def track(
    sequence: Iterable[_T],
    *,
    description: str | bool | None = True,
    record_log_level: int | str | None = "DEBUG",
    report_log_level: int | str | None = "INFO",
    timer: bool = True,
    total: float | None = None,
) -> Generator[_T, Any, None]:
    if description is True:
        description = grapes.caller_location(2)
    description = description or ""
    prog: Progress = progress()
    if timer:
        t = grapes.timer(
            label=description,
            report_at_exit=False,
            record_log_level=record_log_level,
            report_log_level=report_log_level,
        )
        with prog:
            yield from t.track(
                prog.track(sequence, total=total, description=description)
            )
            t.log_report()
    else:
        with prog:
            yield from prog.track(sequence, description=description)



"""Generate the code reference pages and navigation."""

import itertools
from pathlib import Path

import git
from mkdocs_gen_files.editor import FilesEditor
from mkdocs_gen_files.nav import Nav


def git_root() -> Path:
    repo = git.Repo(search_parent_directories=True)
    return Path(repo.working_dir)


def is_private(parts: tuple[str, ...]) -> bool:
    return any(part.startswith("_") for part in parts)


def main() -> None:
    # [Recipes - mkdocstrings](https://mkdocstrings.github.io/recipes/)
    root: Path = git_root()
    editor: FilesEditor = FilesEditor.current()
    nav = Nav()
    src_dir: Path = root / "src"
    for path in sorted(src_dir.rglob("*.py")):
        relative_path: Path = path.relative_to(src_dir)
        doc_path: Path = relative_path.with_suffix(".md")
        parts: tuple[str, ...] = relative_path.with_suffix("").parts
        if parts[-1] == "__init__":
            parts = parts[:-1]
            doc_path = doc_path.with_name("README.md")
        if is_private(parts):
            continue
        nav_parts: tuple[str, ...] = tuple(
            itertools.accumulate(parts, lambda x, y: x + "." + y)
        )
        nav[nav_parts] = doc_path.as_posix()
        full_doc_path = Path("api", doc_path)
        with editor.open(full_doc_path.as_posix(), "w") as fd:
            identifier: str = nav_parts[-1]
            fd.write(f"""::: {identifier}""")
        editor.set_edit_path(
            full_doc_path.as_posix(), (".." / path.relative_to(root)).as_posix()
        )
    with editor.open("api/SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())


main()



import os
import time
from typing import Literal

type CounterName = Literal[
    "monotonic",
    "perf",
    "process",
    "time",
    "thread",
    "user",
    "system",
    "children_user",
    "children_system",
    "elapsed",
]


def get_time(name: CounterName | str = "perf") -> float:  # noqa: C901, PLR0911
    match name.lower():
        case "monotonic":
            return time.monotonic()
        case "perf":
            return time.perf_counter()
        case "process":
            return time.process_time()
        case "time":
            return time.time()
        case "thread":
            return time.thread_time()
        case "user":
            return os.times().user
        case "system":
            return os.times().system
        case "children_user":
            return os.times().children_user
        case "children_system":
            return os.times().children_system
        case "elapsed":
            return os.times().elapsed
        case _:
            msg: str = f"Unsupported time: `{name}`"
            raise ValueError(msg)



import atexit
import contextlib
import functools
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from types import TracebackType
from typing import ParamSpec, Self, TypeVar

from loguru import logger

from liblaf import grapes

from . import TimerRecords, get_time

_P = ParamSpec("_P")
_T = TypeVar("_T")


class timer(Mapping[str, float], contextlib.AbstractContextManager):  # noqa: N801
    counters: Sequence[str]
    depth: int = 0
    label: str | None = None
    records: TimerRecords
    record_log_level: int | str | None = "DEBUG"
    report_log_level: int | str | None = "INFO"
    _start: dict[str, float]
    _end: dict[str, float]

    def __init__(
        self,
        label: str | None = None,
        *,
        counters: Sequence[str] = ["perf", "process"],
        record_log_level: int | str | None = "DEBUG",
        report_at_exit: bool = True,
        report_log_level: int | str | None = "INFO",
    ) -> None:
        self.label = label
        self.counters = counters
        self.record_log_level = record_log_level
        self.report_log_level = report_log_level
        self.records = TimerRecords()
        self._start = {}
        self._end = {}
        if report_at_exit:
            TIMERS.append(self)

    def __getitem__(self, key: str) -> float:
        return self._end[key] - self._start[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.counters)

    def __len__(self) -> int:
        return len(self.counters)

    def __enter__(self) -> Self:
        if self.label is None:
            self.label = grapes.caller_location(2)
        self.depth += 1
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.end()
        self.depth -= 1

    def __call__(self, fn: Callable[_P, _T]) -> Callable[_P, _T]:
        if self.label is None:
            self.label = grapes.full_qual_name(fn)

        @functools.wraps(fn)
        def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            self.depth += 1
            with self:
                ret: _T = fn(*args, **kwargs)
            self.depth -= 1
            return ret

        return wrapped

    def start(self) -> None:
        for time_name in self.counters:
            self._start[time_name] = get_time(time_name)

    def end(self) -> None:
        for name in self.counters:
            self._end[name] = get_time(name)
        self.records.append(self)
        self.depth += 1
        self.log_record()
        self.depth -= 1

    def log_record(self) -> None:
        if not self.record_log_level:
            return
        logger.opt(depth=self.depth + 1).log(self.record_log_level, self.human_record())

    def log_report(self) -> None:
        if not self.report_log_level:
            return
        logger.opt(depth=1).log(self.report_log_level, self.human_report())

    @property
    def elapsed(self) -> float:
        return next(iter(self.values()))

    def human_report(self) -> str:
        return self.records.human_report(self.label)

    def human_record(self) -> str:
        text: str = f"{self.label} > " if self.label else ""
        for k in self.counters:
            human: str = grapes.human_duration(self[k])
            text += f"{k}: {human}, "
        text = text.removesuffix(", ")
        return text

    def track(self, iterable: Iterable[_T]) -> Iterable[_T]:
        self.depth += 1
        if self.label is None:
            self.label = grapes.caller_location(2)
        for item in iterable:
            with self:
                yield item
        self.depth -= 1


TIMERS: list[timer] = []


def exit_hook() -> None:
    for timer in TIMERS:
        if timer.records.count <= 1:
            continue
        timer.log_report()


atexit.register(exit_hook)



import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

from liblaf import grapes

READERS: dict[str, Callable[..., Any]] = {".json": grapes.load_json}
WRITERS: dict[str, Callable[..., None]] = {".json": grapes.save_json}
if grapes.has_module("tomlkit"):
    READERS[".toml"] = grapes.load_toml
    WRITERS[".toml"] = grapes.save_toml
if grapes.has_module("ruamel.yaml"):
    READERS[".yaml"] = grapes.load_yaml
    READERS[".yml"] = grapes.load_yaml
    WRITERS[".yaml"] = grapes.save_yaml
    WRITERS[".yml"] = grapes.save_yaml


def serialize(
    fpath: str | os.PathLike[str], data: Any, *, ext: str | None = None
) -> None:
    fpath: Path = Path(fpath)
    if ext is None:
        ext = fpath.suffix
    if ext not in WRITERS:
        msg: str = f"Unsupported file extension: {ext}"
        raise ValueError(msg)
    writer = WRITERS[ext]
    fpath.parent.mkdir(parents=True, exist_ok=True)
    writer(fpath, data)


def deserialize(fpath: str | os.PathLike[str], *, ext: str | None = None) -> Any:
    fpath: Path = Path(fpath)
    if ext is None:
        ext = fpath.suffix
    if ext not in READERS:
        msg: str = f"Unsupported file extension: {ext}"
        raise ValueError(msg)
    reader = READERS[ext]
    return reader(fpath)



import collections
import textwrap
from collections.abc import Iterator, Mapping

from liblaf import grapes

with grapes.optional_imports(extra="timer"):
    import numpy as np
    import polars as pl


class TimerRecords(Mapping[str, list[float]]):
    _records: dict[str, list[float]]

    def __init__(self) -> None:
        self._records = collections.defaultdict(list)

    def __getitem__(self, key: str) -> list[float]:
        return self._records[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._records)

    def __len__(self) -> int:
        return len(self._records)

    def append(
        self,
        seconds: float | Mapping[str, float] = {},
        nanoseconds: float | Mapping[str, float] = {},
    ) -> None:
        if not isinstance(seconds, Mapping):
            seconds = {"perf": seconds}
        if not isinstance(nanoseconds, Mapping):
            nanoseconds = {"perf": nanoseconds}
        for k, v in seconds.items():
            self._records[k].append(v)
        for k, v in nanoseconds.items():
            self._records[k].append(v * 1e-9)

    @property
    def count(self) -> int:
        return len(next(iter(self._records.values()), []))

    def human_report(self, label: str | None = None) -> str:
        label = label or "Timer"
        text: str = ""
        for k in self._records:
            text += f"{k} > "
            arr: np.ndarray = self.to_numpy(k)
            human_best: str = grapes.human_duration(arr.min())
            human_mean: str = grapes.human_duration_series(arr)
            text += f"mean: {human_mean}, best: {human_best}\n"
        text = text.strip()
        text = f"{label} (total: {self.count})" + "\n" + textwrap.indent(text, "  ")
        return text

    def to_polars(self) -> pl.DataFrame:
        return pl.DataFrame(self._records)

    def to_numpy(self, key: str = "perf") -> np.ndarray:
        return np.asarray(self._records[key])



import math

from liblaf import grapes

with grapes.optional_imports(extra="human"):
    import numpy as np
    import numpy.typing as npt


UNITS: dict[str, float] = {
    "ns": 1e-9,
    "us": 1e-6,
    "ms": 1e-3,
    "s": 1,
    "m": 60,
    "h": 3600,
    "d": 86400,
    "w": 604800,
    "y": 31536000,
}


def get_unit_seconds(unit: str) -> float:
    return UNITS[unit.lower()]


def human_duration_unit_precision(seconds: float) -> tuple[str, int]:  # noqa: C901, PLR0911, PLR0912
    if seconds <= 0:
        return "s", 0
    if seconds < 1e-09:
        return "ns", 3  # .999 ns
    if seconds < 1e-08:
        return "ns", 2  # 9.99 ns
    if seconds < 1e-07:
        return "ns", 1  # 99.9 ns
    if seconds < 1e-06:
        return "ns", 0  # 999. ns
    if seconds < 1e-05:
        return "us", 2  # 9.99 us
    if seconds < 1e-04:
        return "us", 1  # 99.9 us
    if seconds < 1e-03:
        return "us", 0  # 999. us
    if seconds < 1e-02:
        return "ms", 2  # 9.99 ms
    if seconds < 1e-01:
        return "ms", 1  # 99.9 ms
    if seconds < 1:
        return "ms", 0  # 999. ms
    if seconds < 10:
        return "s", 2  # 9.99 s
    if seconds < 60:
        return "s", 1  # 59.9 s
    if seconds < 3600:
        return "m", 0  # 59:59
    if seconds < 86400:
        return "h", 0  # 23:59:59
    return "h", 0


def human_duration(
    seconds: float, unit: str | None = None, precision: int | None = None
) -> str:
    if (unit is None) or (precision is None):
        unit, precision = human_duration_unit_precision(seconds)
    if unit in {"ns", "us", "ms", "s"}:
        unit_seconds: float = get_unit_seconds(unit)
        value: float = seconds / unit_seconds
        human: str = f"{value:.{precision}f}".lstrip("0")
        if precision == 0:
            human += "."
        if unit == "us":
            unit = "µs"
        human += f" {unit}"
        return human
    if unit == "m":
        minutes: int = int(seconds // 60)
        seconds %= 60
        return f"{minutes}:{seconds:02.0f}"
    hours: int = int(seconds // 3600)
    seconds %= 3600
    minutes: int = int(seconds // 60)
    seconds %= 60
    return f"{hours}:{minutes:02.0f}:{seconds:02.0f}"
    # TODO: handle longer durations


def human_duration_with_variance(mean: float, std: float) -> str:
    if not math.isfinite(std):
        return human_duration(mean)
    return human_duration(mean) + " ± " + human_duration(std)


def human_duration_series(series: npt.ArrayLike) -> str:
    series: npt.NDArray = np.asarray(series)
    if series.size <= 1:
        return human_duration(series.item())
    return human_duration_with_variance(series.mean(), series.std())



from collections.abc import Generator


def strip_comments(
    text: str,
    comments: str = "#",
    *,
    strip: bool = True,
    strip_empty_lines: bool = True,
) -> Generator[str, None, None]:
    for raw_line in text.splitlines():
        line: str = raw_line.split(comments, 1)[0]
        if strip:
            line = line.strip()
        if strip_empty_lines and not line:
            continue
        yield line



import json
import os
from pathlib import Path
from typing import Any


def load_json(fpath: str | os.PathLike[str]) -> Any:
    fpath: Path = Path(fpath)
    with fpath.open() as fp:
        return json.load(fp)


def save_json(fpath: str | os.PathLike[str], data: Any) -> None:
    fpath: Path = Path(fpath)
    with fpath.open("w") as fp:
        json.dump(data, fp)



from loguru import logger

from liblaf import grapes


def init_icecream() -> None:
    if not grapes.has_module("icecream"):
        return
    from icecream import ic

    ic.configureOutput(
        prefix="", outputFunction=lambda s: logger.opt(depth=2).log("ICECREAM", s)
    )



import contextlib
from collections.abc import Generator
from typing import Any

import etils.epy


@contextlib.contextmanager
def optional_imports(
    name: str = "liblaf.grapes", extra: str | None = None
) -> Generator[None, Any, None]:
    try:
        yield
    except ImportError as exc:
        suffix: str = f"Missing optional dependency `{exc.name}`."
        if extra is not None:
            suffix += (
                f"\nMake sure to install `{name}` using `pip install {name}[{extra}]`."
            )
        etils.epy.reraise(exc, suffix=suffix)



from . import environ, human, logging, progress_bar, serde, text
from ._optional import has_module, optional_imports
from ._timer import TimerRecords, get_time, timer
from .environ import init_env
from .human import (
    human_count,
    human_duration,
    human_duration_series,
    human_duration_unit_precision,
    human_duration_with_variance,
    human_throughout,
)
from .logging import caller_location, full_qual_name, init_logging, logging_console
from .progress_bar import progress, track
from .serde import (
    deserialize,
    load_json,
    load_pydantic,
    load_toml,
    load_yaml,
    save_json,
    save_pydantic,
    save_toml,
    save_yaml,
    serialize,
)
from .text import strip_comments

__all__ = [
    "TimerRecords",
    "TimerRecords",
    "caller_location",
    "deserialize",
    "environ",
    "full_qual_name",
    "get_time",
    "has_module",
    "human",
    "human_count",
    "human_duration",
    "human_duration_series",
    "human_duration_unit_precision",
    "human_duration_with_variance",
    "human_throughout",
    "init_env",
    "init_logging",
    "load_json",
    "load_pydantic",
    "load_toml",
    "load_yaml",
    "logging",
    "logging_console",
    "optional_imports",
    "progress",
    "progress_bar",
    "save_json",
    "save_pydantic",
    "save_toml",
    "save_yaml",
    "serde",
    "serialize",
    "strip_comments",
    "text",
    "timer",
    "track",
]



from liblaf import grapes

with grapes.optional_imports(extra="duration"):
    import about_time


def human_throughout(value: float, unit: str = "", prec: int | None = None) -> str:
    # TODO: remove dependency on `about-time`
    ht = about_time.HumanThroughput(value, unit)
    return ht.as_human(prec)



import os
from pathlib import Path
from typing import Any

import etils.epy


def callback(exc: ImportError) -> None:
    etils.epy.reraise(
        exc,
        suffix=f"""Missing optional dependency {exc.name}.
Make sure to install liblaf.grapes using `pip install liblaf.grapes[toml]`.""",
    )


with etils.epy.lazy_imports(
    error_callback="Make sure to install `liblaf.grapes` using `pip install liblaf.grapes[toml]`."
):
    import tomlkit


def load_toml(fpath: str | os.PathLike[str]) -> tomlkit.TOMLDocument:
    fpath: Path = Path(fpath)
    with fpath.open() as fp:
        return tomlkit.load(fp)


def save_toml(
    fpath: str | os.PathLike[str], data: Any, *, sort_keys: bool = False
) -> None:
    fpath: Path = Path(fpath)
    with fpath.open("w") as fp:
        tomlkit.dump(data, fp, sort_keys=sort_keys)



from rich.console import Console, RenderableType
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Column
from rich.text import Text

from liblaf import grapes


class RateColumn(ProgressColumn):
    unit: str = "it"

    def __init__(self, unit: str = "it", table_column: Column | None = None) -> None:
        super().__init__(table_column)
        self.unit = unit

    def render(self, task: Task) -> RenderableType:
        if not task.speed:
            return Text(f"?{self.unit}/s", style="progress.data.speed")
        human: str = grapes.human_throughout(task.speed, self.unit)
        return Text(human, style="progress.data.speed")


def progress(
    *columns: str | ProgressColumn, console: Console | None = None
) -> Progress:
    if not columns:
        columns: list[ProgressColumn] = [SpinnerColumn()]
        columns.append(TextColumn("[progress.description]{task.description}"))
        columns += [
            BarColumn(),
            TaskProgressColumn(show_speed=True),
            MofNCompleteColumn(),
            "[",
            TimeElapsedColumn(),
            "<",
            TimeRemainingColumn(),
            ",",
            RateColumn(),
            "]",
        ]
    if not console:
        console = grapes.logging_console()
    progress = Progress(*columns, console=console)
    return progress



import functools

import rich.pretty
import rich.traceback
from rich.console import Console
from rich.style import Style
from rich.theme import Theme


@functools.cache
def logging_theme() -> Theme:
    """...

    References:
        [1] [loguru/loguru/_defaults.py at c490ce0534c6e176306f339a92c221dc6f41a6a7 · Delgan/loguru](https://github.com/Delgan/loguru/blob/c490ce0534c6e176306f339a92c221dc6f41a6a7/loguru/_defaults.py)
    """
    return Theme(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(color="cyan", bold=True),
            "logging.level.debug": Style(color="blue", bold=True),
            "logging.level.icecream": Style(color="magenta", bold=True),
            "logging.level.info": Style(bold=True),
            "logging.level.success": Style(color="green", bold=True),
            "logging.level.warning": Style(color="yellow", bold=True),
            "logging.level.error": Style(color="red", bold=True),
            "logging.level.critical": Style(color="red", bold=True, reverse=True),
        }
    )


@functools.cache
def logging_console() -> Console:
    return rich.console.Console(theme=logging_theme(), stderr=True)


def init_rich(*, show_locals: bool = True) -> None:
    rich.pretty.install(console=logging_console())
    rich.traceback.install(console=logging_console(), show_locals=show_locals)



import contextlib
import inspect
import itertools
import logging
from collections.abc import Iterable, Sequence

import loguru
from environs import Env
from loguru import logger
from rich.logging import RichHandler

from liblaf import grapes

type Filter = "str | loguru.FilterDict | loguru.FilterFunction"


DEFAULT_FILTER: Filter = {
    "": "INFO",
    "__main__": "TRACE",
    "liblaf": "DEBUG",
}


DEFAULT_LEVELS: Sequence["loguru.LevelConfig"] = [
    {"name": "ICECREAM", "no": 15, "color": "<magenta><bold>", "icon": "🍦"}
]


class InterceptHandler(logging.Handler):
    """Intercept standard logging messages toward Loguru sinks.

    References:
        [1] [Overview — loguru documentation](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
    """

    def emit(self, record: logging.LogRecord) -> None:
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def add_level(
    name: str, no: int, color: str | None = None, icon: str | None = None
) -> None:
    with contextlib.suppress(ValueError):
        logger.level(name, no, color=color, icon=icon)


def setup_loguru_logging_intercept(
    level: int | str = logging.NOTSET, modules: Iterable[str] = ()
) -> None:
    """...

    References:
        [1] [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py#L35C5-L42)
    """
    logging.basicConfig(level=level, handlers=[InterceptHandler()])
    for logger_name in itertools.chain(("",), modules):
        mod_logger: logging.Logger = logging.getLogger(logger_name)
        mod_logger.handlers = [InterceptHandler(level=level)]
        mod_logger.propagate = False


def init_loguru(
    level: int | str = logging.NOTSET,
    filter: Filter | None = None,  # noqa: A002
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
) -> None:
    filter = filter or DEFAULT_FILTER  # noqa: A001
    if handlers is None:
        handlers: list[loguru.HandlerConfig] = [
            {
                "sink": RichHandler(
                    console=grapes.logging.logging_console(),
                    omit_repeated_times=False,
                    markup=True,
                    log_time_format="[%Y-%m-%d %H:%M:%S]",
                ),
                "format": "{message}",
                "filter": filter,
            }
        ]
        env: Env = grapes.environ.init_env()
        if fpath := env.path("LOGGING_FILE", None):
            handlers.append({"sink": fpath, "filter": filter, "mode": "w"})
        if fpath := env.path("LOGGING_JSONL", None):
            handlers.append(
                {"sink": fpath, "filter": filter, "serialize": True, "mode": "w"}
            )
    logger.configure(handlers=handlers)
    for lvl in levels or DEFAULT_LEVELS:
        add_level(**lvl)
    setup_loguru_logging_intercept(level=level)
